import dataclasses
import typing as tp
from abc import ABCMeta

import jax
import jax.numpy as jnp
import jax.tree_util
import numpy as np
import typing_extensions as tpe

import treeo as to
from treeo import utils

A = tp.TypeVar("A")
B = tp.TypeVar("B")


class _TrivialPytree:
    def tree_flatten(self):
        tree = vars(self)
        children = (tree,)
        return (children, ())

    @classmethod
    def tree_unflatten(cls, aux, children):
        (tree,) = children

        obj = cls.__new__(cls)
        obj.__dict__.update(tree)

        return obj

    def __init_subclass__(cls):
        jax.tree_util.register_pytree_node_class(cls)


class FieldMetadata:
    node: bool
    kind: type

    def __init__(
        self,
        node: bool,
        kind: type,
    ):
        self.__dict__["node"] = node
        self.__dict__["kind"] = kind

    def update(self, **kwargs) -> "FieldMetadata":
        fields = vars(self).copy()
        fields.update(kwargs)
        return FieldMetadata(**fields)

    def __setattr__(self, name: str, value: tp.Any) -> None:
        raise AttributeError(f"FieldMetadata is immutable, cannot set {name}")


@jax.tree_util.register_pytree_node_class
class Nothing:
    def tree_flatten(self):
        return (), None

    @classmethod
    def tree_unflatten(cls, _aux_data, children):
        return cls()

    def __repr__(self) -> str:
        return "Nothing"

    def __eq__(self, o: object) -> bool:
        return isinstance(o, Nothing)


NOTHING = Nothing()


class Missing:
    pass


MISSING = Missing()


class Hashable(tp.Generic[A]):
    """A hashable immutable wrapper around non-hashable values"""

    value: A

    def __init__(self, value: A):
        self.__dict__["value"] = value

    def __setattr__(self, name: str, value: tp.Any) -> None:
        raise AttributeError(f"Hashable is immutable")



import functools
import inspect
import logging
import re
import typing as tp
from contextlib import contextmanager
from io import StringIO

import jax
import jax.numpy as jnp
import numpy as np
from attr import has

import treeo.tree as tree_m
from treeo import types, utils
from treeo.tree import FieldInfo, FlattenMode, Tree

try:
    from rich.console import Console
    from rich.text import Text
except ImportError:
    Text = None
    Console = None

RICH_WARNING_COUNT = 0

A = tp.TypeVar("A")
B = tp.TypeVar("B")
C = tp.TypeVar("C", bound=tp.Callable)
T = tp.TypeVar("T", bound="Tree")
Filter = tp.Union[
    tp.Type[tp.Any],
    tp.Callable[["FieldInfo"], bool],
]
F = tp.TypeVar("F", bound="tp.Callable")

PAD_START = r"__PAD_START__"
PAD_END = r"__PAD_END__"
FIND_PAD = re.compile(f"{PAD_START}(.*){PAD_END}")
LEAF_TYPES = (types.Nothing, type(None))


# --------------------------------------------------
# functions
# --------------------------------------------------


def filter(
    obj: A,
    *filters: Filter,
    flatten_mode: tp.Union[FlattenMode, str, None] = None,
) -> A:
    """
    The `filter` function allows you to select a subtree by filtering based on a predicate or `kind` type,
    leaves that pass all filters are kept, the rest are set to `Nothing`. For more information see
    [filter's user guide](https://cgarciae.github.io/treeo/user-guide/api/filter).



    Arguments:
        obj: A pytree (possibly containing `to.Tree`s) to be filtered.
        *filters: Types to filter by, membership is determined by `issubclass`, or
            callables that take in a `FieldInfo` and return a `bool`.
        flatten_mode: Sets a new `FlattenMode` context for the operation.
    Returns:
        A new pytree with the filtered fields.

    """

    input_obj = obj

    filters = tuple(
        _get_kind_filter(f) if isinstance(f, tp.Type) else f for f in filters
    )

    def apply_filters(info: tp.Any) -> tp.Any:
        if not isinstance(info, FieldInfo):
            info = FieldInfo(
                name=None,
                value=info,
                kind=type(None),
                module=None,
            )
        assert isinstance(info, FieldInfo)

        return info.value if all(f(info) for f in filters) else types.NOTHING

    with tree_m._CONTEXT.update(add_field_info=True), _flatten_context(flatten_mode):
        obj = jax.tree_map(apply_filters, obj)

    return obj


def merge(
    obj: A,
    other: A,
    *rest: A,
    flatten_mode: tp.Union[FlattenMode, str, None] = None,
    ignore_static: bool = False,
) -> A:
    """
    Creates a new Tree with the same structure but its values merged based on the values from the incoming Trees. For more information see
    [merge's user guide](https://cgarciae.github.io/treeo/user-guide/api/merge).

    Arguments:
        obj: Main pytree to merge.
        other: The pytree first to get the values to merge with.
        *rest: Additional pytree to perform the merge in order from left to right.
        flatten_mode: Sets a new `FlattenMode` context for the operation, if `None` the current context is used. If the current flatten context is `None` and `flatten_mode` is not passed then `FlattenMode.all_fields` is used.
        ignore_static: If `True`, bypasses static fields during the process and the statics fields for output are taken from the first input (`obj`).

    Returns:
        A new pytree with the updated values.
    """

    if flatten_mode is None and tree_m._CONTEXT.flatten_mode is None:
        flatten_mode = FlattenMode.all_fields

    input_obj = obj

    def merge_fn(*xs):
        for x in reversed(xs):
            if not isinstance(x, types.Nothing):
                return x
        return types.NOTHING

    tree_map_fn = _looser_tree_map if ignore_static else jax.tree_map

    with _flatten_context(flatten_mode):
        obj = tree_map_fn(
            merge_fn,
            obj,
            other,
            *rest,
            is_leaf=lambda x: isinstance(x, LEAF_TYPES),
        )

    return obj


def map(
    f: tp.Callable,
    obj: A,
    *filters: Filter,
    flatten_mode: tp.Union[FlattenMode, str, None] = None,
    is_leaf: tp.Callable[[tp.Any], bool] = None,
    field_info: tp.Optional[bool] = False,
) -> A:
    """
    Applies a function to all leaves in a pytree using `jax.tree_map`, if `filters` are given then
    the function will be applied only to the subset of leaves that match the filters. For more information see
    [map's user guide](https://cgarciae.github.io/treeo/user-guide/api/map).


    Arguments:
        f: The function to apply to the leaves.
        obj: a pytree possibly containing `to.Tree`s.
        *filters: The filters used to select the leaves to which the function will be applied.
        flatten_mode: Sets a new `FlattenMode` context for the operation, if `None` the current context is used.
        add_field_info: Represent the leaves of the tree by a `FieldInfo` type. This enables values of the field such as
        kind and value to be used within the `map` function.

    Returns:
        A new pytree with the changes applied.
    """

    input_obj = obj

    has_filters = len(filters) > 0

    with _flatten_context(flatten_mode):
        if has_filters:
            new_obj = filter(obj, *filters)
        else:
            new_obj = obj

        # Conditionally build map function with, or without, the leaf nodes' field info.
        if field_info:
            with add_field_info():
                new_obj: A = jax.tree_map(f, new_obj, is_leaf=is_leaf)
        else:
            new_obj: A = jax.tree_map(f, new_obj, is_leaf=is_leaf)

        if has_filters:
            new_obj = merge(obj, new_obj)

    return new_obj


def to_dict(
    obj: tp.Any,
    *,
    private_fields: bool = False,
    static_fields: bool = True,
    type_info: bool = False,
    field_info: bool = False,
) -> tp.Any:

    if field_info:
        with add_field_info(), flatten_mode(FlattenMode.all_fields):
            flat, treedef = jax.tree_flatten(obj)

        obj = jax.tree_unflatten(treedef, flat)
        obj = tree_m.apply(_remove_field_info_from_metadata, obj)

    return _to_dict(obj, private_fields, static_fields, type_info)


def _remove_field_info_from_metadata(obj: Tree):
    with tree_m._make_mutable_toplevel(obj):
        obj._field_metadata = jax.tree_map(
            lambda x: x.value if isinstance(x, FieldInfo) else x,
            obj._field_metadata,
        )


def _to_dict(
    obj: tp.Any, private_fields: bool, static_fields: bool, type_info: bool
) -> tp.Any:

    if isinstance(obj, Tree):
        fields = vars(obj).copy()

        if not private_fields:
            fields = {k: v for k, v in fields.items() if not k.startswith("_")}

        if not static_fields:
            fields = {k: v for k, v in fields.items() if obj._field_metadata[k].node}

        fields = {
            k: _to_dict(v, private_fields, static_fields, type_info)
            for k, v in fields.items()
        }

        if type_info:
            fields["__type__"] = type(obj)

        return fields
    elif isinstance(obj, tp.Mapping):
        output = {
            k: _to_dict(v, private_fields, static_fields, type_info)
            for k, v in obj.items()
        }

        if type_info:
            output["__type__"] = type(obj)

        return output
    elif isinstance(obj, tp.Sequence) and not isinstance(obj, str):
        output = [_to_dict(v, private_fields, static_fields, type_info) for v in obj]

        if type_info:
            output.append(type(obj))

        return output
    else:
        return obj


def to_string(
    obj: tp.Any,
    private_fields: bool = False,
    static_fields: bool = True,
    color: bool = False,
) -> str:
    """
    Converts a pytree to a string representation.

    Arguments:
        obj: The pytree to convert.
        private_fields: If `True`, private fields are included.
        static_fields: If `True`, static fields are included.

    Returns:
        A string representation of the pytree.
    """
    dict_ = to_dict(
        obj,
        private_fields=private_fields,
        static_fields=static_fields,
        type_info=True,
        field_info=True,
    )
    global RICH_WARNING_COUNT
    rep = _to_string(dict_, level=0, inline=False, color=color, space="    ")
    rep = _add_padding(rep)

    if color:
        if Console is None or Text is None:
            if RICH_WARNING_COUNT < 1:
                RICH_WARNING_COUNT += 1
                logging.warning(
                    f"'rich' library not available, install `rich` to get colors."
                )
        else:
            rep = _get_rich_repr(Text.from_markup(rep))

    return rep


def _to_string(
    obj: tp.Any,
    *,
    level: int,
    inline: bool,
    color: bool,
    space: str,
) -> str:

    indent_level = space * level

    DIM = "[dim]" if color else ""
    END = "[/dim]" if color else ""

    if isinstance(obj, tp.Mapping):
        obj_type = obj["__type__"]
        body = [
            indent_level
            + space
            + f"{field}: {_to_string(value, level=level + 1, inline=True, color=color, space=space)},"
            for field, value in obj.items()
            if field != "__type__"
        ]
        body_str = "\n".join(body)
        type_str = f"{obj_type.__name__}" if inline else obj_type.__name__

        if len(obj) > 1:  # zero fields excluding __type__
            return f"{type_str} {{\n{body_str}\n{indent_level}}}"
        else:
            return f"{type_str} {{}}"

        # return f"\n{body_str}"
    elif isinstance(obj, tp.Sequence) and not isinstance(obj, str):
        obj_type = obj[-1]
        body = [
            indent_level
            + space
            + f"{_to_string(value, level=level + 1, inline=False, color=color, space=space)},"
            for i, value in enumerate(obj[:-1])
        ]
        body_str = "\n".join(body)
        type_str = f"{obj_type.__name__}" if inline else obj_type.__name__

        if len(obj) > 1:  # zero fields excluding __type__
            return f"{type_str} [\n{body_str}\n{indent_level}]"
        else:
            return f"{type_str} []"

    elif isinstance(obj, FieldInfo):
        value = obj.value
        kind_name = obj.kind.__name__ if obj.kind != type(None) else ""

        if isinstance(value, (np.ndarray, jnp.ndarray)):

            value_type = type(value)
            type_module = value_type.__module__.split(".")[0]
            value_rep = (
                f"{type_module}.{value_type.__name__}({value.shape}, {value.dtype})"
            )
        elif isinstance(value, str):
            value_rep = f'"{value}"'
        else:
            value_rep = str(value)

        return (
            f"{value_rep}{PAD_START}{DIM}{kind_name}{END}{PAD_END}"
            if kind_name
            else value_rep
        )

    else:
        return str(obj)


def in_compact() -> bool:
    """
    Returns:
        `True` if current inside a function decorated with `@compact`.
    """
    return tree_m._COMPACT_CONTEXT.in_compact


# ---------------------------------------------------------------
# Context Managers
# ---------------------------------------------------------------


@contextmanager
def add_field_info():
    """
    A context manager that makes `Tree`s produce leaves as `FieldInfo` when flattening.
    """
    with tree_m._CONTEXT.update(add_field_info=True):
        yield


@contextmanager
def flatten_mode(mode: tp.Optional[tp.Union[FlattenMode, str]]):
    """
    A context manager that defines how `Tree`s are flattened. Options are:

    * `'normal'`: Fields are selected as nodes as declared in the class definition (default behavior).
    * `'all_fields'`: All fields are treated as nodes during flattening.
    * `'no_fields'`: All fields are treated as static, `Tree`s produce no leaves.
    * `None`: Context is not changed, current flatten mode is preserved.

    Example:

    ```python
    @dataclass
    class MyTree(Tree):
        x: int # static
        y: int = to.node()

    tree = MyTree(x=1, y=3)

    jax.tree_map(lambda x: x * 2, tree) # MyTree(x=1, y=6)

    with flatten_mode('all_fields'):
        jax.tree_map(lambda x: x + 1, tree) # MyTree(x=2, y=6)
    ```

    Arguments:
        mode: The new flatten mode.
    """
    if mode is not None:
        if isinstance(mode, str):
            mode = FlattenMode(mode)

        with tree_m._CONTEXT.update(flatten_mode=mode):
            yield
    else:
        yield


# alias for internal use
_flatten_context = flatten_mode


# ---------------------------------------------------------------
# decorators
# ---------------------------------------------------------------


def compact(f):
    """
    A decorator that enable the definition of Tree subnodes at runtime.
    """

    if hasattr(f, "_treeo_mutable"):
        raise ValueError(
            f"""Cannot make 'compact' a 'mutable' function, invert the order. If you are using it as a decorator, instead of e.g.
    
    @compact
    @mutable
    def {f.__name__}(self, ...):
    
use:

    @mutable
    @compact
    def {f.__name__}(self, ...):

"""
        )

    @functools.wraps(f)
    def wrapper(tree, *args, **kwargs):
        with tree_m._COMPACT_CONTEXT.compact(f, tree):
            return f(tree, *args, **kwargs)

    wrapper._treeo_compact = True

    return wrapper


def mutable(
    f: tp.Callable[..., A],
    *,
    toplevel_only: bool = False,
) -> tp.Callable[..., tp.Tuple[A, tp.Any]]:
    """
    A decorator that transforms a stateful function `f` that receives an Tree
    instance as a its first argument into a function that returns a tuple of the result and a Tree
    with the new state.

    This is useful for 2 reasons:
    * It transforms `f` into a pure function.
    * It allows `Immutable` Trees to perform inline field updates without getting `RuntimeError`s.

    Note that since the original object is not modified, `Immutable` instance remain in the end immutable.

    Example:

    ```python
    def accumulate_id(tree: MyTree, x: int) -> int:
        tree.n += x
        return x

    tree0 = MyTree(n=4)
    y, tree1 = mutable(accumulate_id)(tree0, 1)

    assert tree0.n == 4
    assert tree1.n == 5
    assert y == 1
    ```

    **Note**: Any `Tree`s that are found in the output of `f` are set to being
    immutable.

    Arguments:
        f: The function to be transformed.
        toplevel_only: If `True`, only the top-level object is made mutable.

    Returns:
        A function that returns a tuple of the result and a Tree with the new state.
    """

    f0 = f

    if inspect.ismethod(f):
        tree0 = f.__self__
        f = f.__func__
    elif isinstance(f, tree_m.Tree) and callable(f):
        tree0 = f
        f = f.__class__.__call__
    else:
        tree0 = None

    if tree0 is not None and not isinstance(tree0, Tree):
        name = f0.__name__ is hasattr(f0, "__name__") and f0.__class__.__name__
        raise TypeError(
            f"Invalid bounded method or callable '{name}', tried to infer unbouded function and instance, "
            f"expected a 'Tree' instance but '{type(tree0).__name__}' instead. Try using an unbounded class method instead."
        )

    @functools.wraps(f)
    def wrapper(tree, *args, **kwargs) -> tp.Tuple[A, tp.Any]:

        tree = tree_m.copy(tree)

        with tree_m.make_mutable(tree, toplevel_only=toplevel_only):
            output = f(tree, *args, **kwargs)

        def _make_output_immutable(a: Tree):
            tree_m._set_mutable(a, None)

        output = tree_m.apply(_make_output_immutable, output)

        return output, tree

    wrapper._treeo_mutable = True

    if tree0 is not None:

        @functools.wraps(f)
        def obj_wrapper(*args, **kwargs):
            return wrapper(tree0, *args, **kwargs)

        return obj_wrapper

    return wrapper


def toplevel_mutable(f: C) -> C:
    """
    A decorator that transforms a stateful function `f` that receives an Tree
    instance as a its first argument into a mutable function. It differs from `mutable`
    in the following ways:

    * It always applies mutability to the top-level object only.
    * `f` is expected to return the new state either as the only output or
        as the last element of a tuple.

    Example:

    ```python
    @dataclass
    class Child(to.Tree, to.Immutable):
        n: int = to.node()

    @dataclass
    def Parent(to.Tree, to.Immutable):
        child: Child

        @to.toplevel_mutable
        def update(self) -> "Parent":
            # self is currently mutable
            self.child = self.child.replace(n=self.child.n + 1) # but child is immutable (so we use replace)

            return self

    tree = Parent(child=Child(n=4))
    tree = tree.update()
    ```

    This behaviour is useful when the top-level tree mostly manipulates sub-trees that have well-defined
    immutable APIs, avoids explicitly run `replace` to propagate updates to the sub-trees and makes
    management of the top-level tree easier.

    **Note**: Any `Tree`s that are found in the output of `f` are set to being
    immutable, however the element is the to the same immutablity status as the
    input tree if they have the same type.

    Arguments:
        f: The function to be transformed.

    Returns:
        A function with top-level mutability.
    """

    f0 = f

    if inspect.ismethod(f):
        tree0 = f.__self__
        f = f.__func__
    elif isinstance(f, tree_m.Tree) and callable(f):
        tree0 = f
        f = f.__class__.__call__
    else:
        tree0 = None

    if tree0 is not None and not isinstance(tree0, Tree):
        name = f0.__name__ is hasattr(f0, "__name__") and f0.__class__.__name__
        raise TypeError(
            f"Invalid bounded method or callable '{name}', tried to infer unbouded function and instance, "
            f"expected a 'Tree' instance but '{type(tree0).__name__}' instead. Try using an unbounded class method instead."
        )

    @functools.wraps(f)
    def wrapper(tree: tree_m.Tree, *args, **kwargs):
        if not isinstance(tree, tree_m.Tree):
            raise TypeError(f"Expected 'Tree' type, got '{type(tree).__name__}'")

        output, _ = mutable(f, toplevel_only=True)(tree, *args, **kwargs)

        if isinstance(output, tuple):
            *ys, last = output
        else:
            ys = ()
            last = output

        if type(last) is type(tree):
            tree_m._set_mutable(last, tree._mutable)

        if isinstance(output, tuple):
            return (*ys, last)
        else:
            return last

    wrapper._treeo_mutable = True

    if tree0 is not None:

        @functools.wraps(f)
        def obj_wrapper(*args, **kwargs):
            return wrapper(tree0, *args, **kwargs)

        return obj_wrapper

    return wrapper


# ---------------------------------------------------------------
# utils
# ---------------------------------------------------------------


def _looser_tree_map(
    f: tp.Callable[..., tp.Any],
    tree: tp.Any,
    *rest: tp.Any,
    is_leaf: tp.Optional[tp.Callable[[tp.Any], bool]] = None,
) -> tp.Any:
    jax.tree_map
    leaves, treedef = jax.tree_flatten(
        tree,
        is_leaf=is_leaf,
    )
    all_leaves = [leaves] + [jax.tree_flatten(r, is_leaf=is_leaf)[0] for r in rest]

    n_leaves = len(leaves)
    assert all(len(l) == n_leaves for l in all_leaves)

    return treedef.unflatten(f(*xs) for xs in zip(*all_leaves))


def _get_kind_filter(
    t: type,
) -> tp.Callable[[FieldInfo], bool]:
    def _filter(info: FieldInfo) -> bool:
        return (
            info.kind is not None
            and isinstance(t, tp.Type)
            and issubclass(info.kind, t)
        )

    return _filter


def _get_rich_repr(table):
    assert Console is not None
    f = StringIO()
    console = Console(file=f, force_terminal=True)
    console.print(table)

    return f.getvalue()


def _add_padding(text):

    space = " "
    lines = text.split("\n")
    padded_info = []
    for i in range(len(lines)):
        match = FIND_PAD.search(lines[i])
        if match:
            lines[i] = FIND_PAD.sub("", lines[i])
            padded_info.append(match[1])
        else:
            padded_info.append("")
    lenghts = [len(line) for line in lines]
    max_length = max(lenghts)

    text = "\n".join(
        line + space * (max_length - length) + f"    {info}" if info else line
        for line, length, info in zip(lines, lenghts, padded_info)
    )

    return text



import dataclasses
import inspect
import re
import typing as tp

import jax
import jax.numpy as jnp
import numpy as np
import typing_extensions as tpe

A = tp.TypeVar("A")
B = tp.TypeVar("B")

_pymap = map
_pyfilter = filter


@tpe.runtime_checkable
class ArrayLike(tpe.Protocol):
    shape: tp.Tuple[int, ...]
    dtype: np.dtype


def field(
    default: tp.Any = dataclasses.MISSING,
    *,
    node: bool,
    kind: type = type(None),
    default_factory: tp.Any = dataclasses.MISSING,
    init: bool = True,
    repr: bool = True,
    hash: tp.Optional[bool] = None,
    compare: bool = True,
) -> tp.Any:

    return dataclasses.field(
        default=default,
        metadata={
            "node": node,
            "kind": kind,
        },
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
    )


def node(
    default: tp.Any = dataclasses.MISSING,
    *,
    kind: type = type(None),
    default_factory: tp.Any = dataclasses.MISSING,
    init: bool = True,
    repr: bool = True,
    hash: tp.Optional[bool] = None,
    compare: bool = True,
) -> tp.Any:
    return field(
        default=default,
        node=True,
        kind=kind,
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
    )


def static(
    default: tp.Any = dataclasses.MISSING,
    *,
    kind: type = type(None),
    default_factory: tp.Any = dataclasses.MISSING,
    init: bool = True,
    repr: bool = True,
    hash: tp.Optional[bool] = None,
    compare: bool = True,
) -> tp.Any:
    return field(
        default,
        node=False,
        kind=kind,
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
    )


def _get_all_annotations(cls: type) -> tp.Dict[str, type]:
    d = {}
    for c in reversed(cls.mro()):
        if hasattr(c, "__annotations__"):
            d.update(**c.__annotations__)
    return d


def _get_all_vars(cls: type) -> tp.Dict[str, tp.Any]:
    d = {}
    for c in reversed(cls.mro()):
        if hasattr(c, "__dict__"):
            d.update(vars(c))
    return d


def _all_types(t: type) -> tp.Iterable[type]:
    return _pyfilter(lambda t: isinstance(t, tp.Type), _all_types_unfiltered(t))


def _all_types_unfiltered(t: type) -> tp.Iterable[type]:
    yield t

    if hasattr(t, "__args__"):
        yield t.__origin__

        for arg in t.__args__:
            yield from _all_types_unfiltered(arg)


def _unique_name(
    names: tp.Set[str],
    name: str,
):

    if name in names:

        match = re.match(r"(.*?)(\d*)$", name)
        assert match is not None

        name = match[1]
        num_part = match[2]

        i = int(num_part) if num_part else 2
        str_template = f"{{name}}{{i:0{len(num_part)}}}"

        while str_template.format(name=name, i=i) in names:
            i += 1

        name = str_template.format(name=name, i=i)

    names.add(name)
    return name


def _unique_names(
    names: tp.Iterable[str],
    *,
    existing_names: tp.Optional[tp.Set[str]] = None,
) -> tp.Iterable[str]:
    if existing_names is None:
        existing_names = set()

    for name in names:
        yield _unique_name(existing_names, name)


def _lower_snake_case(s: str) -> str:
    s = re.sub(r"(?<!^)(?=[A-Z])", "_", s).lower()
    parts = s.split("_")
    output_parts = []

    for i in range(len(parts)):
        if i == 0 or len(parts[i - 1]) > 1:
            output_parts.append(parts[i])
        else:
            output_parts[-1] += parts[i]

    return "_".join(output_parts)


def _get_name(obj) -> str:
    if hasattr(obj, "name") and obj.name:
        return obj.name
    elif hasattr(obj, "__name__") and obj.__name__:
        return _lower_snake_case(obj.__name__)
    elif hasattr(obj, "__class__") and obj.__class__.__name__:
        return _lower_snake_case(obj.__class__.__name__)
    else:
        raise ValueError(f"Could not get name for: {obj}")


def _safe_update_fields_from(obj, updates):
    for field, value in updates.__dict__.items():
        setattr(obj, field, value)
    return obj


def _get_unbound_method(obj: object, method: tp.Union[str, tp.Callable]) -> tp.Callable:

    if isinstance(method, str) and not hasattr(obj, method):
        raise TypeError(f"class '{type(obj).__name__}' has no attribute {method}")

    return (
        getattr(obj.__class__, method)
        if isinstance(method, str)
        else method.__func__
        if inspect.ismethod(method)
        else method
    )



# isort:skip_file
__version__ = "0.1.2"


from treeo.api import (
    filter,
    merge,
    map,
    to_dict,
    in_compact,
    add_field_info,
    flatten_mode,
    to_string,
    compact,
    mutable,
    toplevel_mutable,
)
from treeo.mixins import (
    Copy,
    ToString,
    ToDict,
    Repr,
    Filter,
    Merge,
    Map,
    Apply,
    Compact,
    Extensions,
    KindMixin,
    Immutable,
    ImmutableTree,
    MutabilityError,
)
from treeo.tree import FlattenMode, FieldInfo, TreeMeta, Tree, copy, apply, make_mutable
from treeo.types import FieldMetadata, Nothing, NOTHING, Missing, MISSING, Hashable
from treeo.utils import ArrayLike, field, node, static


__all__ = [
    "Apply",
    "ArrayLike",
    "Compact",
    "Copy",
    "Extensions",
    "FieldInfo",
    "FieldMetadata",
    "Filter",
    "FlattenMode",
    "Hashable",
    "Immutable",
    "ImmutableTree",
    "KindMixin",
    "MISSING",
    "Map",
    "Merge",
    "Missing",
    "NOTHING",
    "Nothing",
    "Repr",
    "ToDict",
    "ToString",
    "Tree",
    "TreeMeta",
    "add_field_info",
    "apply",
    "compact",
    "copy",
    "field",
    "filter",
    "flatten_mode",
    "in_compact",
    "map",
    "merge",
    "mutable",
    "toplevel_mutable",
    "node",
    "static",
    "to_dict",
    "to_string",
    "make_mutable",
    "MutabilityError",
]



import dataclasses
import functools
import inspect
import threading
import typing as tp
from contextlib import contextmanager

from treeo import api
from treeo import tree as tree_m
from treeo import types, utils

A = tp.TypeVar("A")
C = tp.TypeVar("C", bound=tp.Callable)


class Copy:
    """
    Mixin that adds a `.copy()` method to the class.
    """

    def copy(self: A) -> A:
        """
        `copy` is a wrapper over `treeo.copy` that passes `self` as the first argument.
        """
        return tree_m.copy(self)


class ToString:
    """
    Mixin that adds a `.to_string()` method to the class.
    """

    def to_string(
        self: A,
        *,
        private_fields: bool = False,
        static_fields: bool = True,
        color: bool = False,
    ) -> str:
        """
        `to_string` is a wrapper over `treeo.to_string` that passes `self` as the first argument.

        Arguments:
            private_fields: If `True`, private fields are included.
            static_fields: If `True`, static fields are included.
            color: If `True`, color is included.

        Returns:
            A string representation of the object.
        """
        return api.to_string(
            self,
            private_fields=private_fields,
            static_fields=static_fields,
            color=color,
        )


class ToDict:
    """
    Mixin that adds a `.to_dict()` method to the class.
    """

    def to_dict(
        self,
        *,
        private_fields: bool = False,
        static_fields: bool = True,
        type_info: bool = False,
        field_info: bool = False,
    ) -> tp.Any:
        """
        `to_dict` is a wrapper over `treeo.to_dict` that passes `self` as the first argument.

        Arguments:
            private_fields: If `True`, private fields are included.
            static_fields: If `True`, static fields are included.
            type_info: If `True`, type information is included.
            field_info: If `True`, field information is included.

        Returns:
            A dict representation of the object.
        """
        return api.to_dict(
            self,
            private_fields=private_fields,
            static_fields=static_fields,
            type_info=type_info,
            field_info=field_info,
        )


class Repr:
    """
    Mixin that adds a `__repr__` method to the class.
    """

    def __repr__(self) -> tp.Any:
        """
        Uses `treeo.to_string` to generate a string representation of the object.
        """
        return api.to_string(
            self,
            private_fields=False,
            static_fields=True,
            color=False,
        )


class Filter:
    """
    Mixin that adds a `.filter()` method to the class.
    """

    def filter(
        self: A,
        *filters: api.Filter,
        flatten_mode: tp.Union[api.FlattenMode, str, None] = None,
    ) -> A:
        """
        `filter` is a wrapper over `treeo.filter` that passes `self` as the first argument.

        Arguments:
            *filters: Types to filter by, membership is determined by `issubclass`, or
                callables that take in a `FieldInfo` and return a `bool`.
            flatten_mode: Sets a new `FlattenMode` context for the operation.

        Returns:
            A new pytree with the filtered fields.
        """
        return api.filter(self, *filters, flatten_mode=flatten_mode)


class Merge:
    """
    Mixin that adds a `.merge()` method to the class.
    """

    def merge(
        self: A,
        other: A,
        *rest: A,
        flatten_mode: tp.Union[api.FlattenMode, str, None] = None,
        ignore_static: bool = False,
    ) -> A:
        """
        `merge` is a wrapper over `treeo.merge` that passes `self` as the first argument.

        Arguments:
            other: The pytree first to get the values to merge with.
            *rest: Additional pytree to perform the merge in order from left to right.
            flatten_mode: Sets a new `FlattenMode` context for the operation, if `None` the current context is used. If the current flatten context is `None` and `flatten_mode` is not passed then `FlattenMode.all_fields` is used.
            ignore_static: If `True`, bypasses static fields during the process and the statics fields for output are taken from the first input (`obj`).

        Returns:
            A new pytree with the merged values.
        """
        return api.merge(
            self,
            other,
            *rest,
            flatten_mode=flatten_mode,
            ignore_static=ignore_static,
        )


class Map:
    """
    Mixin that adds a `.map()` method to the class.
    """

    def map(
        self: A,
        f: tp.Callable,
        *filters: api.Filter,
        flatten_mode: tp.Union[api.FlattenMode, str, None] = None,
        is_leaf: tp.Callable[[tp.Any], bool] = None,
    ) -> A:
        """
        `map` is a wrapper over `treeo.map` that passes `self` as the second argument.

        Arguments:
            f: The function to apply to the leaves.
            *filters: The filters used to select the leaves to which the function will be applied.
            flatten_mode: Sets a new `FlattenMode` context for the operation, if `None` the current context is used.

        Returns:
            A new pytree with the changes applied.
        """
        return api.map(
            f,
            self,
            *filters,
            flatten_mode=flatten_mode,
            is_leaf=is_leaf,
        )


class Apply:
    """
    Mixin that adds a `.apply()` method to the class.
    """

    def apply(self: A, f: tp.Callable[..., None], *rest: A, inplace: bool = False) -> A:
        """
        `apply` is a wrapper over `treeo.apply` that passes `self` as the second argument.

        Arguments:
            f: The function to apply.
            *rest: additional pytrees.
            inplace: If `True`, the input `obj` is mutated.

        Returns:
            A new pytree with the updated Trees or the same input `obj` if `inplace` is `True`.
        """
        return tree_m.apply(f, self, *rest, inplace=inplace)


class Compact:
    _field_metadata: tp.Dict[str, types.FieldMetadata]
    _subtrees: tp.Optional[tp.Tuple[str, ...]]

    @property
    def first_run(self) -> bool:
        """
        Returns:
            `True` if its currently the first run of a `compact` method.
        """
        if tree_m._COMPACT_CONTEXT.current_tree is not self:
            raise RuntimeError(
                f"Object '{type(self).__name__}' is not the current tree, found '{type(tree_m._COMPACT_CONTEXT.current_tree).__name__}', did you forget the @compact decorator?"
            )

        return self._subtrees is None

    # NOTE: it feels like `get_field` could be safely used in non-`compact` methods, maybe
    # the various checks done to verify that this method is used inside `compact` could be removed.
    def get_field(
        self,
        field_name: str,
        initializer: tp.Callable[[], A],
    ) -> A:
        """
        A method that gets a field with the given name if exists, otherwise it initializes it and returns it.

        Currently the follow restrictions apply:

        * The field must be declared in the class definition.
        * The method can only be called inside a `compact` context.

        Arguments:
            field_name: The name of the field to get.
            initializer: The function to initialize the field if it does not exist.

        Returns:
            The field value.
        """
        value: A

        if field_name not in self._field_metadata:
            raise ValueError(f"Metadata for field '{field_name}' does not exist.")

        if field_name in vars(self):
            value = getattr(self, field_name)
        else:
            if tree_m._COMPACT_CONTEXT.in_compact and not self.first_run:
                raise RuntimeError(
                    f"Trying to initialize field '{field_name}' after the first run of `compact`."
                )

            value = initializer()

            with tree_m._make_mutable_toplevel(self):
                setattr(self, field_name, value)

        return value


class Extensions(Copy, ToString, ToDict, Repr, Filter, Merge, Map, Apply, Compact):
    """
    Mixin that adds all available mixins from `treeo.mixins` except `KindMixin`.
    """

    pass


class KindMixin:
    @classmethod
    def field(
        cls,
        default: tp.Any = dataclasses.MISSING,
        *,
        node: bool,
        default_factory: tp.Any = dataclasses.MISSING,
        init: bool = True,
        repr: bool = True,
        hash: tp.Optional[bool] = None,
        compare: bool = True,
    ) -> tp.Any:
        return utils.field(
            default=default,
            node=node,
            kind=cls,
            default_factory=default_factory,
            init=init,
            repr=repr,
            hash=hash,
            compare=compare,
        )

    @classmethod
    def node(
        cls,
        default: tp.Any = dataclasses.MISSING,
        *,
        default_factory: tp.Any = dataclasses.MISSING,
        init: bool = True,
        repr: bool = True,
        hash: tp.Optional[bool] = None,
        compare: bool = True,
    ) -> tp.Any:
        return utils.node(
            default=default,
            kind=cls,
            default_factory=default_factory,
            init=init,
            repr=repr,
            hash=hash,
            compare=compare,
        )

    @classmethod
    def static(
        cls,
        default: tp.Any = dataclasses.MISSING,
        default_factory: tp.Any = dataclasses.MISSING,
        init: bool = True,
        repr: bool = True,
        hash: tp.Optional[bool] = None,
        compare: bool = True,
    ) -> tp.Any:
        return cls.field(
            default=default,
            node=False,
            default_factory=default_factory,
            init=init,
            repr=repr,
            hash=hash,
            compare=compare,
        )


# -------------------------------------------------------------------------------
# Immutable Mixin
# -------------------------------------------------------------------------------


class Immutable:
    """
    Mixin that makes a class immutable. It adds a `.replace()` and `.mutable()` methods to the class
    which let you modify the state by creating a new objects.
    """

    _mutable: tp.Optional[tree_m._MutableState]

    @property
    def is_mutable(self) -> bool:
        """
        Returns:
            `True` if the object is mutable.
        """
        return self._mutable is not None

    def replace(self: A, **kwargs) -> A:
        """
        Returns a copy of the Tree with the given fields specified in `kwargs`
        updated to the new values.

        Example:

        ```python
        @dataclass
        class MyTree(to.Tree, to.Immutable):
            x: int = to.node()

        tree = MyTree(x=1)

        # increment x by 1
        tree = tree.replace(x=tree.x + 1)
        ```

        Arguments:
            **kwargs: The fields to update.

        Returns:
            A new Tree with the updated fields.
        """
        tree: tree_m.Tree = tree_m.copy(self)

        with tree_m._make_mutable_toplevel(tree):
            for key, value in kwargs.items():
                setattr(tree, key, value)

        tree._update_local_metadata()
        # return a copy to potentially update metadata
        return tree


class MutabilityError(AttributeError):
    """
    Raised when an operation is attempted on an immutable object.
    """

    pass


# define __setattr__ outside of class so linters still detect it unknown attribute assignments
def _immutable_setattr(self: Immutable, key: str, value: tp.Any) -> None:
    if not self._mutable:
        raise MutabilityError(
            f"Trying to mutate field '{key}' in immutable '{type(self).__name__}' object."
        )

    object.__setattr__(self, key, value)


Immutable.__setattr__ = _immutable_setattr


class ImmutableTree(tree_m.Tree, Immutable):
    """
    A Tree class that also inherits from `Immutable`.
    """

    pass



import dataclasses
import enum
import functools
import inspect
import threading
import typing as tp
from abc import ABCMeta
from contextlib import contextmanager
from copy import copy as python_copy
from dataclasses import dataclass
from types import MappingProxyType

import jax
import jax.tree_util

from treeo import types, utils

T = tp.TypeVar("T", bound="Tree")
A = tp.TypeVar("A")


class FlattenMode(enum.Enum):
    no_fields = enum.auto()
    normal = enum.auto()
    all_fields = enum.auto()


@dataclass
class _Context(threading.local):
    add_field_info: bool = False
    flatten_mode: tp.Optional[FlattenMode] = None

    def __enter__(self):
        global _CONTEXT
        self._old_context = _CONTEXT
        _CONTEXT = self

    def __exit__(self, *args):
        global _CONTEXT
        _CONTEXT = self._old_context

    @contextmanager
    def update(self, **kwargs):
        fields = vars(self).copy()
        fields.pop("_old_context", None)
        fields.update(kwargs)

        with _Context(**fields):
            yield


@dataclass
class _CompactContext(threading.local):
    current_tree: tp.Optional["Tree"] = None
    existing_subtrees: tp.Optional[tp.List["Tree"]] = None
    tree_idx: int = 0
    new_subtrees: tp.Optional[tp.List["Tree"]] = None
    compact_calls: tp.Optional[tp.Set[tp.Callable[..., tp.Any]]] = None

    @property
    def in_compact(self):
        return self.existing_subtrees is not None or self.new_subtrees is not None

    def __enter__(self):
        global _COMPACT_CONTEXT
        self._old_context = _COMPACT_CONTEXT
        _COMPACT_CONTEXT = self

    def __exit__(self, *args):
        global _COMPACT_CONTEXT
        _COMPACT_CONTEXT = self._old_context

    @contextmanager
    def update(self, **kwargs):
        fields = vars(self).copy()
        fields.pop("_old_context", None)
        fields.update(kwargs)

        with _CompactContext(**fields):
            yield

    @contextmanager
    def compact(self, f: tp.Callable[..., tp.Any], tree: "Tree"):

        if self.current_tree is tree:
            assert self.compact_calls is not None

            if f in self.compact_calls:
                raise RuntimeError(f"Detected recursion in `compact` for: {tree}")

            self.compact_calls.add(f)
            yield
            return

        new_subtrees: tp.Optional[tp.List["Tree"]]
        existing_subtrees: tp.Optional[tp.List["Tree"]]

        if tree._subtrees is None:
            existing_subtrees = None
            new_subtrees = []
        else:
            existing_subtrees = list(getattr(tree, field) for field in tree._subtrees)
            new_subtrees = None

        with _CompactContext(
            current_tree=tree,
            existing_subtrees=existing_subtrees,
            new_subtrees=new_subtrees,
            compact_calls={f},
        ):
            yield
            assert _COMPACT_CONTEXT.current_tree is not None
            tree = _COMPACT_CONTEXT.current_tree
            existing_subtrees = _COMPACT_CONTEXT.existing_subtrees
            new_subtrees = _COMPACT_CONTEXT.new_subtrees

        if tree._subtrees is None:
            with _make_mutable_toplevel(tree):
                assert new_subtrees is not None
                field_names = list(
                    utils._unique_names(
                        (utils._get_name(new_tree) for new_tree in new_subtrees),
                        existing_names=set(vars(tree).keys()),
                    )
                )
                tree._subtrees = tuple(field_names)
                for field, subtree in zip(field_names, new_subtrees):
                    if (
                        field in tree._field_metadata
                        and not tree._field_metadata[field].node
                    ):
                        raise ValueError(
                            f"Trying to subtree '{type(subtree).__name__}' to field '{field}' of '{type(tree).__name__}' but it has previously been declared with `node=False`"
                        )

                    if field not in tree._field_metadata:
                        tree._field_metadata[field] = types.FieldMetadata(
                            kind=type(None),
                            node=True,
                        )

                    setattr(tree, field, subtree)


@dataclass
class _MutableContext(threading.local):
    prev_mutable: tp.Optional[tp.Dict[int, tp.Optional["_MutableState"]]] = None

    def __enter__(self):
        global _MUTABLE_CONTEXT
        self._old_context = _MUTABLE_CONTEXT
        _MUTABLE_CONTEXT = self

    def __exit__(self, *args):
        global _MUTABLE_CONTEXT
        _MUTABLE_CONTEXT = self._old_context

    @contextmanager
    def update(self, **kwargs):
        fields = vars(self).copy()
        fields.pop("_old_context", None)
        fields.update(kwargs)

        with _MutableContext(**fields):
            yield


_COMPACT_CONTEXT = _CompactContext()
_CONTEXT = _Context()
_MUTABLE_CONTEXT = _MutableContext()


class FieldInfo:
    def __init__(
        self,
        name: tp.Optional[str],
        value: tp.Any,
        kind: tp.Type[tp.Any],
        module: tp.Optional["Tree"],
    ):
        self.name = name
        self.value = value
        self.kind = kind
        self.module = module

    def __repr__(self) -> str:
        return f"FieldInfo(name={self.name!r}, value={self.value!r}, kind={self.kind!r}, module={type(self.module).__name__})"


class TreeMeta(ABCMeta):
    def __call__(cls: tp.Type[T], *args, **kwargs) -> T:
        obj: T

        if _COMPACT_CONTEXT.existing_subtrees is not None:
            if len(_COMPACT_CONTEXT.existing_subtrees) <= _COMPACT_CONTEXT.tree_idx:
                raise ValueError(
                    f"Out of bounds: trying to new initialize new Tree '{cls.__name__}' after the first compact run"
                )

            obj = _COMPACT_CONTEXT.existing_subtrees[_COMPACT_CONTEXT.tree_idx]
            _COMPACT_CONTEXT.tree_idx += 1

            return obj
        else:
            obj = cls.__new__(cls)
            with _make_mutable_toplevel(obj):
                obj = cls.construct(obj, *args, **kwargs)

        if _COMPACT_CONTEXT.new_subtrees is not None:
            _COMPACT_CONTEXT.new_subtrees.append(obj)

        if (
            _COMPACT_CONTEXT.in_compact
            and _COMPACT_CONTEXT.current_tree is not None
            and _COMPACT_CONTEXT.current_tree._mutable == _MutableState.ALL
        ):
            _set_mutable(obj, _COMPACT_CONTEXT.current_tree._mutable)

            if _MUTABLE_CONTEXT.prev_mutable is not None:
                _MUTABLE_CONTEXT.prev_mutable[id(obj)] = None

        return obj

    def construct(cls, obj: T, *args, **kwargs) -> T:

        obj._field_metadata = obj._field_metadata.copy()

        # set default fields
        for field, default_factory in obj._factory_fields.items():
            setattr(obj, field, default_factory())

        for field, default_value in obj._default_field_values.items():
            setattr(obj, field, default_value)

        # reset context before __init__ and add obj as current tree
        with _CompactContext(current_tree=obj):
            obj.__init__(*args, **kwargs)

        # auto-annotations
        obj._update_local_metadata()

        return obj


class _MutableState(enum.Enum):
    TOPLEVEL = 1
    ALL = 2


class Tree(metaclass=TreeMeta):
    _field_metadata: tp.Dict[str, types.FieldMetadata]
    _factory_fields: tp.Dict[str, tp.Callable[[], tp.Any]]
    _default_field_values: tp.Dict[str, tp.Any]
    _subtrees: tp.Optional[tp.Tuple[str, ...]]
    _mutable: tp.Optional[_MutableState]

    @property
    def field_metadata(self) -> tp.Mapping[str, types.FieldMetadata]:
        return MappingProxyType(self._field_metadata)

    def update_field_metadata(
        self: T,
        field: str,
        node: tp.Optional[bool] = None,
        kind: tp.Optional[type] = None,
    ) -> T:
        module = copy(self)

        field_metadata = module._field_metadata[field]
        updates = {}

        if node is not None:
            updates.update(node=node)

        if kind is not None:
            updates.update(kind=kind)

        if updates:
            field_metadata = field_metadata.update(**updates)

        module._field_metadata[field] = field_metadata

        return module

    def check_metadata_updates(self):
        """
        Checks for new fields, if found, adds them to the metadata.
        """
        # flattening / unflattening process updates metadata
        updates = copy(self)
        utils._safe_update_fields_from(self, updates)

    def _update_local_metadata(self):
        for field, value in vars(self).items():

            if field not in self._field_metadata:
                self._field_metadata[field] = types.FieldMetadata(
                    node=isinstance(value, Tree),
                    kind=type(value),
                )

    def __init_subclass__(cls):
        jax.tree_util.register_pytree_node_class(cls)

        # Restore the signature
        sig = inspect.signature(cls.__init__)
        parameters = tuple(sig.parameters.values())
        cls.__signature__ = sig.replace(parameters=parameters[1:])

        annotations = utils._get_all_annotations(cls)
        class_vars = utils._get_all_vars(cls)

        # init class variables
        cls._field_metadata = {}
        cls._factory_fields = {}
        cls._default_field_values = {}
        cls._subtrees = None
        cls._mutable = None

        for field, value in class_vars.items():
            if isinstance(value, dataclasses.Field):

                # save defaults
                if value.default is not dataclasses.MISSING:
                    cls._default_field_values[field] = value.default
                elif value.default_factory is not dataclasses.MISSING:
                    cls._factory_fields[field] = value.default_factory

                # extract metadata
                if value.metadata is not None and "node" in value.metadata:
                    cls._field_metadata[field] = types.FieldMetadata(
                        node=value.metadata["node"],
                        kind=value.metadata["kind"],
                    )

        for field, value in annotations.items():

            if field not in cls._field_metadata:
                is_node = any(issubclass(t, Tree) for t in utils._all_types(value))
                cls._field_metadata[field] = types.FieldMetadata(
                    node=is_node,
                    kind=type(None),
                )

    def tree_flatten(self):

        tree = python_copy(self)

        fields = vars(tree).copy()

        node_fields = {}
        static_fields = {}

        # auto-annotations
        tree._update_local_metadata()

        if _CONTEXT.flatten_mode == FlattenMode.all_fields:
            node_fields = fields
        elif _CONTEXT.flatten_mode == FlattenMode.no_fields:
            static_fields = fields
        else:  # normal or None
            for field, value in fields.items():
                field_annotation = tree.field_metadata[field]

                if field_annotation.node:
                    node_fields[field] = value
                else:
                    static_fields[field] = value

        # maybe convert to FieldInfo
        if _CONTEXT.add_field_info:
            for field in node_fields.keys():
                if field in TREE_PRIVATE_FIELDS:
                    continue

                kind = tree._field_metadata[field].kind
                # leaves, treedef
                node_fields[field] = jax.tree_map(
                    lambda x: FieldInfo(
                        name=field,
                        value=x,
                        kind=kind,
                        module=tree,
                    )
                    if not isinstance(x, Tree)
                    else x,
                    node_fields[field],
                    is_leaf=lambda x: isinstance(x, Tree),
                )

        children = (node_fields,)

        return children, static_fields

    @classmethod
    def tree_unflatten(cls, static_fields, children):

        module = cls.__new__(cls)
        (node_fields,) = children

        if _CONTEXT.add_field_info:
            for field in node_fields.keys():
                node_fields[field] = jax.tree_map(
                    lambda x: x.value if isinstance(x, FieldInfo) else x,
                    node_fields[field],
                    is_leaf=lambda x: isinstance(x, Tree),
                )

        module.__dict__.update(node_fields, **static_fields)

        return module


TREE_PRIVATE_FIELDS = {
    x for x in list(vars(Tree)) + list(Tree.__annotations__) if x.startswith("_")
}


def copy(obj: A, shallow: bool = False) -> A:
    """
    Returns a deep copy of the tree, almost equivalent to:
    ```python
    jax.tree_map(lambda x: x, self)
    ```
    but will try to copy static nodes as well.
    """
    if shallow:
        isleaf = lambda x: isinstance(x, Tree) and x is not obj
    else:
        isleaf = None

    with _CONTEXT.update(flatten_mode=FlattenMode.all_fields):
        return jax.tree_map(lambda x: x, obj, is_leaf=isleaf)


def apply(
    f: tp.Callable[..., None],
    obj: A,
    *rest: A,
    inplace: bool = False,
    _top_inplace: tp.Optional[bool] = None,
    _top_level: bool = True,
) -> A:
    """
    Applies a function to all `to.Tree`s in a Pytree. Works very similar to `jax.tree_map`,
    but its values are `to.Tree`s instead of leaves, also `f` should apply the changes inplace to Tree object.

    Arguments:
        f: The function to apply.
        obj: a pytree possibly containing Trees.
        *rest: additional pytrees.
        inplace: If `True`, the input `obj` is mutated.

    Returns:
        A new pytree with the updated Trees or the same input `obj` if `inplace` is `True`.
    """
    if _top_inplace is None:
        _top_inplace = inplace

    if _top_level:
        rest = copy(rest)
        if not inplace:
            obj = copy(obj)

    objs = (obj,) + rest

    def nested_fn(obj, *rest):
        if isinstance(obj, Tree):
            apply(
                f,
                obj,
                *rest,
                inplace=True,
                _top_inplace=_top_inplace,
                _top_level=False,
            )

    jax.tree_map(
        nested_fn,
        *objs,
        is_leaf=lambda x: isinstance(x, Tree) and not x in objs,
    )

    if isinstance(obj, Tree):
        if _top_inplace or obj._mutable:
            f(obj, *rest)
        else:
            with _make_mutable_toplevel(obj):
                f(obj, *rest)

    return obj


def make_mutable(tree: Tree, toplevel_only: bool = False):
    """
    Returns a context manager that makes a tree mutable.

    Arguments:
        tree: The tree to make mutable.
        toplevel_only: If `True`, only the top-level tree is made mutable.
    """
    if toplevel_only:
        return _make_mutable_toplevel(tree)
    else:
        return _make_mutable(tree)


def _make_mutable_fn(tree: Tree):
    assert _MUTABLE_CONTEXT.prev_mutable is not None
    _MUTABLE_CONTEXT.prev_mutable[id(tree)] = tree._mutable
    # update __dict__ instead to avoid error during __setattr__
    _set_mutable(tree, _MutableState.ALL)


def _revert_mutable_fn(tree: Tree):
    assert _MUTABLE_CONTEXT.prev_mutable is not None
    # update __dict__ instead to avoid error during __setattr__
    obj_id = id(tree)
    if obj_id in _MUTABLE_CONTEXT.prev_mutable:
        _set_mutable(tree, _MUTABLE_CONTEXT.prev_mutable[obj_id])
    else:
        _set_mutable(tree, None)


@contextmanager
def _make_mutable(obj):
    """
    Context manager that makes the tree mutable.
    """

    with _MUTABLE_CONTEXT.update(prev_mutable={}):
        try:
            apply(_make_mutable_fn, obj, inplace=True)
            yield
        finally:
            apply(_revert_mutable_fn, obj, inplace=True)


@contextmanager
def _make_mutable_toplevel(tree):
    """
    Context manager that makes a single tree mutable.
    """
    _mutable = tree._mutable

    _set_mutable(tree, _MutableState.TOPLEVEL)

    try:
        yield
    finally:
        # update __dict__ instead to avoid error during __setattr__
        _set_mutable(tree, _mutable)


def _set_mutable(obj: Tree, value: tp.Optional[_MutableState]):
    obj.__dict__["_mutable"] = value



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['treeo']

package_data = \
{'': ['*']}

install_requires = \
['typing-extensions>=3.7.0']

setup_kwargs = {
    'name': 'treeo',
    'version': '0.2.0.dev0',
    'description': '',
    'long_description': '# Treeo\n\n_A small library for creating and manipulating custom JAX Pytree classes_\n\n* **Light-weight**: has no dependencies other than `jax`.\n* **Compatible**: Treeo `Tree` objects are compatible with any `jax` function that accepts Pytrees.\n* **Standards-based**: `treeo.field` is built on top of python\'s `dataclasses.field`.\n* **Flexible**: Treeo is compatible with both dataclass and non-dataclass classes.\n\nTreeo lets you easily create class-based Pytrees so your custom objects can easily interact seamlessly with JAX. Uses of Treeo can range from just creating simple simple JAX-aware utility classes to using it as the core abstraction for full-blown frameworks. Treeo was originally extracted from the core of [Treex](https://github.com/cgarciae/treex) and shares a lot in common with [flax.struct](https://flax.readthedocs.io/en/latest/flax.struct.html#module-flax.struct).\n\n[Documentation](https://cgarciae.github.io/treeo) | [User Guide](https://cgarciae.github.io/treeo/user-guide/intro)\n\n## Installation\nInstall using pip:\n```bash\npip install treeo\n```\n\n## Basics\nWith Treeo you can easily define your own custom Pytree classes by inheriting from Treeo\'s `Tree` class and using the `field` function to declare which fields are nodes (children) and which are static (metadata):\n\n```python\nimport treeo as to\n\n@dataclass\nclass Person(to.Tree):\n    height: jnp.array = to.field(node=True) # I am a node field!\n    name: str = to.field(node=False) # I am a static field!\n```\n`field` is just a wrapper around `dataclasses.field` so you can define your Pytrees as dataclasses, but Treeo fully supports non-dataclass classes as well. Since all `Tree` instances are Pytree they work with the various functions from the`jax` library as expected:\n\n```python\np = Person(height=jnp.array(1.8), name="John")\n\n# Trees can be jitted!\njax.jit(lambda person: person)(p) # Person(height=array(1.8), name=\'John\')\n\n# Trees can be mapped!\njax.tree_map(lambda x: 2 * x, p) # Person(height=array(3.6), name=\'John\')\n```\n#### Kinds\nTreeo also include a kind system that lets you give semantic meaning to fields (what a field represents within your application). A kind is just a type you pass to `field` via its `kind` argument: \n\n```python\nclass Parameter: pass\nclass BatchStat: pass\n\nclass BatchNorm(to.Tree):\n    scale: jnp.ndarray = to.field(node=True, kind=Parameter)\n    mean: jnp.ndarray = to.field(node=True, kind=BatchStat)\n```\n\nKinds are very useful as a filtering mechanism via [treeo.filter](https://cgarciae.github.io/treeo/user-guide/api/filter):\n\n```python \nmodel = BatchNorm(...)\n\n# select only Parameters, mean is filtered out\nparams = to.filter(model, Parameter) # BatchNorm(scale=array(...), mean=Nothing)\n```\n`Nothing` behaves like `None` in Python, but it is a special value that is used to represent the absence of a value within Treeo.\n\nTreeo also offers the [merge](https://cgarciae.github.io/treeo/user-guide/api/merge) function which lets you rejoin filtered Trees with a logic similar to Python `dict.update` but done recursively:\n```python hl_lines="3"\ndef loss_fn(params, model, ...):\n    # add traced params to model\n    model = to.merge(model, params)\n    ...\n\n# gradient only w.r.t. params\nparams = to.filter(model, Parameter) # BatchNorm(scale=array(...), mean=Nothing)\ngrads = jax.grad(loss_fn)(params, model, ...)\n```\n\nFor a more in-depth tour check out the [User Guide](https://cgarciae.github.io/treeo/user-guide/intro).\n\n## Examples\n\n### A simple Tree\n```python\nfrom dataclasses import dataclass\nimport treeo as to\n\n@dataclass\nclass Character(to.Tree):\n    position: jnp.ndarray = to.field(node=True)    # node field\n    name: str = to.field(node=False, opaque=True)  # static field\n\ncharacter = Character(position=jnp.array([0, 0]), name=\'Adam\')\n\n# character can freely pass through jit\n@jax.jit\ndef update(character: Character, velocity, dt) -> Character:\n    character.position += velocity * dt\n    return character\n\ncharacter = update(character velocity=jnp.array([1.0, 0.2]), dt=0.1)\n```\n### A Stateful Tree\n```python\nfrom dataclasses import dataclass\nimport treeo as to\n\n@dataclass\nclass Counter(to.Tree):\n    n: jnp.array = to.field(default=jnp.array(0), node=True) # node\n    step: int = to.field(default=1, node=False) # static\n\n    def inc(self):\n        self.n += self.step\n\ncounter = Counter(step=2) # Counter(n=jnp.array(0), step=2)\n\n@jax.jit\ndef update(counter: Counter):\n    counter.inc()\n    return counter\n\ncounter = update(counter) # Counter(n=jnp.array(2), step=2)\n\n# map over the tree\n```\n\n### Full Example - Linear Regression\n\n```python\nimport jax\nimport jax.numpy as jnp\nimport matplotlib.pyplot as plt\nimport numpy as np\n\nimport treeo as to\n\n\nclass Linear(to.Tree):\n    w: jnp.ndarray = to.node()\n    b: jnp.ndarray = to.node()\n\n    def __init__(self, din, dout, key):\n        self.w = jax.random.uniform(key, shape=(din, dout))\n        self.b = jnp.zeros(shape=(dout,))\n\n    def __call__(self, x):\n        return jnp.dot(x, self.w) + self.b\n\n\n@jax.value_and_grad\ndef loss_fn(model, x, y):\n    y_pred = model(x)\n    loss = jnp.mean((y_pred - y) ** 2)\n\n    return loss\n\n\ndef sgd(param, grad):\n    return param - 0.1 * grad\n\n\n@jax.jit\ndef train_step(model, x, y):\n    loss, grads = loss_fn(model, x, y)\n    model = jax.tree_map(sgd, model, grads)\n\n    return loss, model\n\n\nx = np.random.uniform(size=(500, 1))\ny = 1.4 * x - 0.3 + np.random.normal(scale=0.1, size=(500, 1))\n\nkey = jax.random.PRNGKey(0)\nmodel = Linear(1, 1, key=key)\n\nfor step in range(1000):\n    loss, model = train_step(model, x, y)\n    if step % 100 == 0:\n        print(f"loss: {loss:.4f}")\n\nX_test = np.linspace(x.min(), x.max(), 100)[:, None]\ny_pred = model(X_test)\n\nplt.scatter(x, y, c="k", label="data")\nplt.plot(X_test, y_pred, c="b", linewidth=2, label="prediction")\nplt.legend()\nplt.show()\n```\n',
    'author': 'Cristian Garcia',
    'author_email': 'cgarcia.e88@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://cgarciae.github.io/treeo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
